@@ -32,6 +32,8 @@
 (pushnew :berkeley-db *features*)
 (in-package "BDB")
 
+(setf (documentation (find-package "BDB") 'sys::impnotes) "berkeley-db")
+
 ;;; objects
 (cl:defstruct (bdb-handle (:constructor nil) (:copier nil))
   (handle nil :read-only t)
@@ -345,7 +347,7 @@
 )
 
 (define-condition bdb-error (simple-error)
-  (($errno :initarg :errno :reader bdb-error-number)))
+  (($ecode :reader bdb-error-code :initarg :code)))
 
 ;;; restore locks
 (pushnew "BDB" custom:*system-package-list* :test #'string=)
