@@ -3,14 +3,12 @@
 (in-package "EXT")
 (export
  '(custom::*prompt-start* custom::*prompt-step* custom::*prompt-break*
-   custom::*prompt-body* custom::*prompt-finish*)
+   custom::*prompt-body* custom::*prompt-finish* custom::*user-commands*)
  "CUSTOM")
+(ext:re-export "CUSTOM" "EXT")
 (export
- '(*command-index* prompt-new-package package-short-name
-   #+unix make-xterm-io-stream
-   break-level step-level
-   custom::*prompt-start* custom::*prompt-step* custom::*prompt-break*
-   custom::*prompt-body* custom::*prompt-finish*)
+ '(*command-index* prompt-new-package
+   break-level step-level)
  "EXT")
 (in-package "SYSTEM")
 
@@ -31,21 +29,6 @@
 ;; Counter to avoid infinite recursion due to *debug-io*
 (defvar *recurse-count-debug-io* 0)
 
-
-;; Returns the shortest (nick)name of the package.
-(defun package-short-name (pkg)
-  (declare (type package pkg))
-  (let ((name (reduce (lambda (st0 st1)
-                        (declare (simple-string st0 st1))
-                        (if (> (length st0) (length st1)) st1 st0))
-                      (package-nicknames pkg) :initial-value
-                      (package-name pkg))))
-    (case *print-case*
-      (:upcase (string-upcase name))
-      (:downcase (string-downcase name))
-      (:capitalize (string-capitalize name))
-      (t name))))
-
 ; The number of commands received so far.
 (defvar *command-index* 0)
 
@@ -88,7 +71,8 @@
         (format nil "~@[~a~][~:d]"
                 (if (or (not (find-symbol "T" *package*))
                         (prompt-new-package))
-                    (package-short-name *package*))
+                    ;; use symbol so that ~A will respect *PRINT-CASE*
+                    (make-symbol (package-shortest-name *package*)))
                 (incf *command-index*))
         (TEXT "[*package* invalid]")))
   "The main top level prompt.")
@@ -111,12 +95,13 @@
 (defun debug-reset-io ()
   (rotatef *package* *saved-debug-package*)
   (rotatef *readtable* *saved-debug-readtable*)
-  (format *debug-io* (TEXT "~&Reset *PACKAGE* to ~s") *package*)
+  (fresh-line *debug-io*)
+  (format *debug-io* (TEXT "Reset *PACKAGE* to ~s") *package*)
   (throw 'debug 'continue))
 
 ;; Components of the Break-Loop:
 (defvar *debug-frame*)
-(defvar *debug-mode*)
+(defvar *debug-mode* 4)
 ; lower bound for frame-down/frame-down-1
 (defvar *frame-limit1* nil)
 ; upper bound for frame-up and frame-up-1
@@ -176,13 +161,16 @@
 (defun get-frame-limit ()
   (let (number)
     (loop
-     (write-string (TEXT "
-Enter the limit for max. frames to print or ':all' for all: ") *debug-io*)
+     (fresh-line *debug-io*)
+     (write-string (TEXT "Enter the limit for max. frames to print or ':all' for all: ") *debug-io*)
      (setq number (read-from-string (read-line *debug-io* nil nil)))
      (if (or (integerp number) (eq number :all))
        (return)
-       (format *debug-io* (TEXT "~&~A is not a number. Try again.")
-               number)))
+       (progn
+         (fresh-line *debug-io*)
+         (format *debug-io* (TEXT "~A is not a number. Try again.")
+                 number)
+         (elastic-newline *debug-io*))))
     (unless (eq number :all)
       number)))
 
@@ -191,16 +179,25 @@ Enter the limit for max. frames to print or ':all' for all: ") *debug-io*)
   (setq *debug-print-frame-limit* (get-frame-limit))
   (throw 'debug 'continue))
 
+(defun print-backtrace (&key ((:out *standard-output*) *standard-output*)
+                        (mode *debug-mode*) (limit *debug-print-frame-limit*))
+  ;; SHOW-STACK prints its output to *STANDARD-OUTPUT*, so we bind that
+  (let ((frame-count
+         (show-stack
+          mode limit
+          (frame-down-1 (frame-up-1 (or *frame-limit1* (frame-limit1 13)) mode)
+                        mode))))
+    (fresh-line *standard-output*)
+    (format *standard-output* (TEXT "Printed ~D frames") frame-count)
+    (elastic-newline *standard-output*)))
+
 ;;; debug-backtrace with-optional 'print-limit'
 (defun debug-backtrace (&optional (mode *debug-mode*)
                                   (limit *debug-print-frame-limit*)
                                   (prompt-limit-p nil))
-  (let ((frame-count
-         (show-stack mode       ; local limit takes precedence:
-                     (or (and prompt-limit-p (get-frame-limit)) limit)
-                     (frame-down-1 (frame-up-1 *frame-limit1* mode) mode))))
-    (format *debug-io* (TEXT "~&Printed ~D frames") frame-count)
-    (throw 'debug 'continue)))
+  (print-backtrace :out *debug-io* :mode mode
+                   :limit (or (and prompt-limit-p (get-frame-limit)) limit))
+  (throw 'debug 'continue))
 
 (defun debug-backtrace-1 () (debug-backtrace 1))
 (defun debug-backtrace-2 () (debug-backtrace 2))
@@ -234,19 +231,38 @@ Enter the limit for max. frames to print or ':all' for all: ") *debug-io*)
 
 ;;; print it
 (defun print-error (condition)
-  (format *debug-io* (TEXT "~%Last error is: >> ~A <<")
-          (print-condition condition nil)))
+  (fresh-line *debug-io*)
+  (write-string (TEXT "The last error:") *debug-io*)
+  (pretty-print-condition condition *debug-io* :text-indent 3)
+  (elastic-newline *debug-io*))
+
+(defvar *user-commands* nil
+  "The list of functions, each of which should return a list of bindings.
+A `binding' is either a doc string (printed by `Help' or `:h')
+or a pair (STRING . FUNCTION) so that typing STRING will call FUNCTION.")
+
+(defun wrap-user-commands (functions)
+  "wrap user commands in THROWs"
+  (mapcar (lambda (binding)
+            (etypecase binding
+              (string binding)
+              (cons (cons (car binding)
+                          (lambda ()
+                            (funcall (cdr binding))
+                            (throw 'debug 'continue))))))
+          (mapcap #'funcall functions)))
 
 ;; extended commands
 (defun commands0 ()
-  (list
-   (TEXT "
+  (list*
+   (TEXT "You are in the top-level Read-Eval-Print loop.
 Help (abbreviated :h) = this list
 Use the usual editing capabilities.
 \(quit) or (exit) leaves CLISP.")
 
    (cons "Help"         #'debug-help)
-   (cons ":h"           #'debug-help)))
+   (cons ":h"           #'debug-help)
+   (wrap-user-commands *user-commands*)))
 
 (defun commands1 ()
   (list
@@ -389,7 +405,8 @@ Continue       :c       switch off single step mode, continue evaluation
            (commands2))
          (when may-continue
            (commands3))
-         commandsr))
+         commandsr
+         (wrap-user-commands *user-commands*)))
 
 ;; Main-Loop with additional help-command
 (defun main-loop ()
@@ -397,16 +414,18 @@ Continue       :c       switch off single step mode, continue evaluation
   (driver                 ; build driver-frame; do #'lambda "infinitely"
    #'(lambda ()
        (catch 'debug            ; catch the (throw 'debug ...)
-         (when (read-eval-print   ; read-eval-print INPUT-line
-                (string-concat (prompt-start) (prompt-body) (prompt-finish))
-                (commands0))
-           ;; T -> #<EOF>
-           ;; NIL -> form is already evaluated
-           ;;        result has been printed
-           (if (interactive-stream-p *standard-input*)
-             (exit)           ; user typed EOF
-             (progn (setq *command-index* 0) ; reset *command-index*
-                    (return-from main-loop)))))))) ; and proceed
+         ;; ANSI CL wants an ABORT restart to be available.
+         (with-restarts ((ABORT () (throw 'debug 'continue)))
+           (when (read-eval-print   ; read-eval-print INPUT-line
+                   (string-concat (prompt-start) (prompt-body) (prompt-finish))
+                   (commands0))
+             ;; T -> #<EOF>
+             ;; NIL -> form is already evaluated
+             ;;        result has been printed
+             (if (interactive-stream-p *standard-input*)
+               (exit)           ; user typed EOF
+               (progn (setq *command-index* 0) ; reset *command-index*
+                      (return-from main-loop))))))))) ; and proceed
 
 (setq *driver* #'main-loop)
 
@@ -414,7 +433,15 @@ Continue       :c       switch off single step mode, continue evaluation
                    &aux
                    (may-continue
                     (or continuable
-                        (and condition (find-restart 'continue condition))))
+                        (and condition
+                             (let ((restart
+                                     (find-restart 'CONTINUE condition)))
+                               (and restart
+                                    ;; Ignore the CONTINUE restart if it is not
+                                    ;; useful without prior corrective action,
+                                    ;; otherwise it leads to user frustration.
+                                    (restart-meaningfulp restart)
+                                    restart)))))
                    (interactive-p (interactive-stream-p *debug-io*))
                    (commandsr '()))
   (when (and print-it (typep condition (clos:find-class 'condition)))
@@ -435,7 +462,8 @@ Continue       :c       switch off single step mode, continue evaluation
 
     (if may-continue
       (progn
-        (write-string "** - Continuable Error" *error-output*)
+        (write-string "** - " *error-output*)
+        (write-string (TEXT "Continuable Error") *error-output*)
         (terpri *error-output*))
       (write-string "*** - " *error-output*))
 
@@ -446,7 +474,8 @@ Continue       :c       switch off single step mode, continue evaluation
           (setq *recursive-error-count* 0)
           (write-string (TEXT "Unprintable error message.")
                         *error-output*))
-        (sys::print-condition condition *error-output*)))
+        (pretty-print-condition condition *error-output*
+                                :text-indent (if may-continue 5 6))))
 
     ;; Now the error message is on the screen; give the user some information
     ;; how to continue from continuable errors.
@@ -454,15 +483,17 @@ Continue       :c       switch off single step mode, continue evaluation
     (when may-continue
       (if continuable
         (when interactive-p
-          (terpri *debug-io*)
+          (fresh-line *debug-io*)
           (write-string (TEXT "You can continue (by typing 'continue').")
-                        *debug-io*))
+                        *debug-io*)
+          (elastic-newline *debug-io*))
         (progn
-          (terpri *debug-io*)
+          (fresh-line *debug-io*)
           (when interactive-p
             (write-string (TEXT "If you continue (by typing 'continue'): ")
                           *debug-io*))
-          (princ may-continue *debug-io*)))))
+          (princ may-continue *debug-io*)
+          (elastic-newline *debug-io*)))))
 
   (when condition
     (let ((restarts (remove may-continue (compute-restarts condition)))
@@ -471,25 +502,27 @@ Continue       :c       switch off single step mode, continue evaluation
                            (TEXT "The following restarts are available:"))))
       (when restarts
         (when interactive-p
-          (terpri *debug-io*)
-          (write-string restarts-help *debug-io*))
+          (fresh-line *debug-io*)
+          (write-string restarts-help *debug-io*)
+          (elastic-newline *debug-io*))
         (let ((counter 0))
           (dolist (restart restarts)
             (let* ((command
                     (string-concat ":R" (sys::decimal-string (incf counter))))
                    (name (string (restart-name restart)))
-                   (helpstring (format nil "~%~A~15T~A~24T~A" name command
+                   (helpstring (format nil "~A~15T~A~24T~A" name command
                                        (princ-to-string restart)))
                    (restart restart)  ; for FUNC
                    (func #'(lambda () (invoke-restart-interactively restart))))
               ;; display the restarts:
               (when interactive-p
-                (write-string helpstring *debug-io*))
-              (push helpstring commandsr)
+                (fresh-line *debug-io*)
+                (write-string helpstring *debug-io*)
+                (elastic-newline *debug-io*))
+              (push (string-concat (string #\Newline) helpstring) commandsr)
               ;; put it into the commandsr list.
               (push (cons command func) commandsr)
               (push (cons name func) commandsr)))
-          (terpri *debug-io*)
           (setq commandsr (cons (string-concat (string #\Newline) restarts-help)
                                 (nreverse commandsr)))))))
   (force-output *debug-io*)
@@ -499,7 +532,6 @@ Continue       :c       switch off single step mode, continue evaluation
     (makunbound '*terminal-read-open-object*)
     (clear-input *debug-io*) ; because the user did not expect a break loop
     (let* ((*break-count* (1+ *break-count*))
-           (*fasoutput-stream* nil) ; compiled code may be called right away
            (stream (make-synonym-stream '*debug-io*))
            (*standard-input* stream)
            (*standard-output* stream)
@@ -507,7 +539,7 @@ Continue       :c       switch off single step mode, continue evaluation
                                   (prompt-body) (prompt-finish)))
            (*frame-limit1* (frame-limit1 13))
            (*frame-limit2* (frame-limit2))
-           (*debug-mode* 4)
+           (*debug-mode* *debug-mode*)
            (*debug-frame*
             (frame-down-1 (frame-up-1 *frame-limit1* *debug-mode*)
                           *debug-mode*))
@@ -516,16 +548,18 @@ Continue       :c       switch off single step mode, continue evaluation
        ;; build driver frame and repeat #'lambda (infinitely; ...)
        #'(lambda ()
            (case (catch 'debug    ; catch (throw 'debug ...) and analyse
-                   ;; build environment *debug-frame*
-                   ;; which is valid/equal for/to *debug-frame*
-                   (same-env-as *debug-frame*
-                     #'(lambda ()
-                         (if (read-eval-print ; read-eval-print INPUT-line
-                              prompt commands-list)
+                   ;; ANSI CL wants an ABORT restart to be available.
+                   (with-restarts ((ABORT () (throw 'debug 'continue)))
+                     ;; build environment *debug-frame*
+                     ;; which is valid/equal for/to *debug-frame*
+                     (same-env-as *debug-frame*
+                       #'(lambda ()
+                           (if (read-eval-print ; read-eval-print INPUT-line
+                                 prompt commands-list)
                              ;; T -> #<EOF>
                              ;; NIL -> form is already evaluated;
                              ;;        result has been printed
-                             (throw 'debug (if may-continue 'quit 'unwind))))))
+                             (throw 'debug (if may-continue 'quit 'unwind)))))))
              ((print-error) (print-error condition))
              ((inspect-error) (inspect condition))
              ((unwind) (go unwind))
@@ -545,11 +579,11 @@ Continue       :c       switch off single step mode, continue evaluation
 ;;;--------------------------------------------------------------------------
 ;;;        convenient Stepper. (runs only if compiled!)
 
-(defvar *step-level* 0 "current Step-depth")
-(defvar *step-quit* most-positive-fixnum "critical Step-depth")
+(defvar *step-level* 0 "current Step-depth") ; ABI
+(defvar *step-quit* most-positive-fixnum "critical Step-depth") ; ABI
 ;; the stepper wakes up, as soon as *step-level* <= *step-quit*
 
-(defvar *step-watch* nil)               ; terminating condition
+(defvar *step-watch* nil)               ; terminating condition ; ABI
 
 (defmacro step (form)
   "(STEP form), CLTL p. 441"
@@ -561,7 +595,7 @@ Continue       :c       switch off single step mode, continue evaluation
 
 (defun step-values (values)
   (let ((*standard-output* *debug-io*))
-    (terpri #|*debug-io*|#)
+    (fresh-line #|*debug-io*|#)
     (write-string (TEXT "step ") #|*debug-io*|#)
     (write *step-level* #|:stream *debug-io*|#)
     (write-string " ==> " #|*debug-io*|#)
@@ -574,7 +608,8 @@ Continue       :c       switch off single step mode, continue evaluation
          (do ((L values))
              ((endp L))
            (write (pop L) #|:stream *debug-io*|#)
-           (unless (endp L) (write-string ", " #|*debug-io*|#))))))
+           (unless (endp L) (write-string ", " #|*debug-io*|#)))))
+    (elastic-newline #|*debug-io*|#))
   (values-list values))
 
 (defun step-level () *step-level*)
@@ -598,7 +633,7 @@ Continue       :c       switch off single step mode, continue evaluation
                                     (prompt-body) (prompt-finish)))
              (*frame-limit1* (frame-limit1 11))
              (*frame-limit2* (frame-limit2))
-             (*debug-mode* 4)
+             (*debug-mode* *debug-mode*)
              (*debug-frame*
               (frame-down-1 (frame-up-1 *frame-limit1* *debug-mode*)
                             *debug-mode*))
@@ -615,21 +650,23 @@ Continue       :c       switch off single step mode, continue evaluation
                 (driver
                   ;;  build driver frame and repeat #'lambda (infinitely ...)
                   #'(lambda ()
+                      ;; catch the (throw 'debug ...) and analyse
                       (case
                           (catch 'debug
-                            ;; catch the (throw 'debug ...) and analyse
-                            (same-env-as *debug-frame*
-                              ;; build environment *debug-frame* that
-                              ;; is valid/equal for/to *debug-frame*
-                              #'(lambda ()
-                                  (if (read-eval-print ; get/read INPUT-line
-                                       prompt commands-list)
+                            ;; ANSI CL wants an ABORT restart to be available.
+                            (with-restarts ((ABORT () (throw 'debug 'continue)))
+                              (same-env-as *debug-frame*
+                                ;; build environment *debug-frame* that
+                                ;; is valid/equal for/to *debug-frame*
+                                #'(lambda ()
+                                    (if (read-eval-print ; get/read INPUT-line
+                                          prompt commands-list)
                                       ;; T -> #<EOF>
                                       (go continue)
                                       ;; NIL -> form is already evaluated;
                                       ;;        result has been printed
-                                    #|(throw 'debug 'continue)|#
-                                      ))))
+                                      #|(throw 'debug 'continue)|#
+                                      )))))
                         ((unwind) (go unwind))
                         ((abort-to-top) (go abort-to-top))
                         (t ))))) ; other cases, especially continue
@@ -660,33 +697,6 @@ Continue       :c       switch off single step mode, continue evaluation
       (go over))))
 
 ;;;--------------------------------------------------------------------------
-#+unix
-(defun make-xterm-io-stream ()
-  (let* ((pipe (with-open-stream (s (make-pipe-input-stream
-                                     "mktemp /tmp/clisp-x-io-XXXXXX"))
-                 (read-line s)))
-         (title "CLISP I/O")
-         tty-name xio
-         (clos::*warn-if-gf-already-called* nil))
-    (shell (format nil "rm -f ~a; mknod ~a p; ~
-                   xterm -n ~s -T ~s -e 'tty >> ~a; cat ~a' &"
-                   pipe pipe title title pipe pipe))
-    (setq tty-name
-          (with-open-file (s pipe :direction :input) (read-line s))
-          xio (make-two-way-stream
-               (open tty-name :direction :input)
-               (open tty-name :direction :output)))
-    (defmethod close :after ((x (eql xio)) &rest junk)
-      (declare (ignore x junk))
-      (with-open-file (s pipe :direction :output)
-        (write-line (TEXT "Bye.") s))
-      (delete-file pipe)
-      (close (two-way-stream-input-stream xio))
-      (close (two-way-stream-output-stream xio))
-      (let ((clos::*warn-if-gf-already-called* nil))
-        (remove-method #'close (find-method #'close '(:after) `((eql ,xio))))))
-    xio))
-;;;--------------------------------------------------------------------------
 
 ;; Now that condition.lisp is loaded and *break-driver* has a value:
 ;; Activate the Condition System.
